﻿using System;
using System.ServiceModel;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NorthwindEntities;
using Microsoft.ComplexEventProcessing;
using Microsoft.ComplexEventProcessing.Linq;
using Microsoft.ComplexEventProcessing.ManagementService;

namespace NorthwindEntities
{
    class Program
    {
        static void Main(string[] args)
        {
            // Connect to the Northwind sample database on SQL Server. Use the default Entity Model
            // generated by the Entity Framework for the Northwind sample database.

            using (NorthwindEntities northwind = new NorthwindEntities())
            {
                Server server = Server.Create("Default");
                Application application = server.CreateApplication("app");

                ServiceHost host = new ServiceHost(server.CreateManagementService());

                host.AddServiceEndpoint(
                    typeof(IManagementService),
                    new WSHttpBinding(SecurityMode.Message),
                    "http://localhost/MyStreamInsightServer");
                
                // To enable remote connection / debugging, you also need to uncomment the
                // lines "host.Open()" and "host.Close()".
                // In addition, the URI needs to be provisioned for the
                // account that runs this process (unless Administrator). To do this, open
                // an admin shell and execute the following command, using the respective
                // user credentials:
                // netsh http add urlacl url=http://+:80/MyStreamInsightServer user=<domain\username>

                host.Open();

                // Query all Orders where there is a known order date, ship date and ship region.
                var databaseQuery = from o in northwind.Orders
                                    where o.OrderDate.HasValue && o.ShippedDate.HasValue && o.ShipRegion != null
                                    orderby o.OrderDate.Value
                                    select o;

                // Transform the query results into a stream of interval events whose start and end 
                // times are defined by the order and ship timestamps. Keep track of the shipping region.
                var streamSource = databaseQuery.ToIntervalStream(application, 
                         o => IntervalEvent.CreateInsert(
                             o.OrderDate.Value,
                             o.ShippedDate.Value,
                             new { o.ShipRegion }), AdvanceTimeSettings.IncreasingStartTime);

                // Find time intervals during which more than 3 orders are in process within a region.
                var streamQuery = from o in streamSource
                                  group o by o.ShipRegion into g
                                  from window in g.SnapshotWindow(SnapshotWindowOutputPolicy.Clip)
                                  select new { OrderCount = window.Count(), ShipRegion = g.Key } into agg
                                  where agg.OrderCount > 2
                                  select agg;

                // Convert temporal query results into an enumerable result of interval events. This example
                // filters out CTI events, and projects the relevant portions of the interval event.
                var results = from intervalEvent in streamQuery.ToIntervalEnumerable()
                              where intervalEvent.EventKind != EventKind.Cti
                              select new
                              {
                                  intervalEvent.StartTime,
                                  intervalEvent.EndTime,
                                  intervalEvent.Payload.OrderCount,
                                  intervalEvent.Payload.ShipRegion,
                              };

                //Console.WriteLine("Press enter to finish");
                //Console.ReadLine();

                // Enumerating the results triggers the underlying SQL Server and StreamInsight queries.
                foreach (var activeInterval in results)
                {
                    Console.WriteLine("Between {0} and {1}, {2} orders were active in the '{3}' region.",
                        activeInterval.StartTime,
                        activeInterval.EndTime,
                        activeInterval.OrderCount,
                        activeInterval.ShipRegion);
                    
                    Console.WriteLine("Press enter to continue");
                    Console.ReadLine();
                }
            }
        }
    }
}
