﻿namespace StreamInsight.Samples.ExplicitServer
{
    using System;
    using System.ServiceModel;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Threading;
    using Microsoft.ComplexEventProcessing;
    using Microsoft.ComplexEventProcessing.Linq;
    using Microsoft.ComplexEventProcessing.ManagementService;
    using StreamInsight.Samples.Extensibility.UserFunctions;
    using StreamInsight.Samples.InputAdapters.TextFile;
    using StreamInsight.Samples.OutputAdapters.TextFile;
    
    /// <summary>
    /// Console application that demonstrates the StreamInsight object model API.
    /// </summary>
    public class ExplicitServer
    {
        public static void Main(string[] args)
        {
            // Create a tracer to output information on the console.
            TraceListener tracer = new ConsoleTraceListener();

            // Create the CEP server in-process.
            //
            // In order to connect to an existing server instead of creating a
            // new one, Server.Connect() can be used, with the appropriate
            // endpoint exposed by the remote server (see documentation of the
            // Server class or the README.txt for this sample for more
            // information).
            tracer.WriteLine("Creating CEP Server");

            // BB: You must name your server
            Server server = Server.Create("default");

            try
            {
                // Create application in server. The application will serve 
                // as a container for actual CEP objects and queries.
                tracer.WriteLine("Creating CEP Application");
                Application application = server.CreateApplication("ObjectModelSample");

                // Create query logic as a query template
                tracer.WriteLine("Registering LINQ query template");
                QueryTemplate queryTemplate = CreateQueryTemplate(application);

                // Register adapter factories, so that the engine will be able 
                // to instantiate adapters when the query starts.
                tracer.WriteLine("Registering Adapter Factories");
                InputAdapter csvInputAdapter = application.CreateInputAdapter<TextFileInputFactory>("CSV Input", "Reading tuples from a CSV file");
                OutputAdapter csvOutputAdapter = application.CreateOutputAdapter<TextFileOutputFactory>("CSV Output", "Writing result events to a CSV file");

                // bind query to event producers and consumers
                QueryBinder queryBinder = BindQuery(csvInputAdapter, csvOutputAdapter, queryTemplate);

                // Create bound query that can be run
                tracer.WriteLine("Registering bound query");

                // BB: Switch arguments
                Query query = application.CreateQuery("TrafficSensorQuery", "Minute average count, filtered by location threshold", queryBinder);

                tracer.WriteLine("Start query");

                // In this sample, we use text files as input event source. 
                // The application will stop as soon as all events have been
                // consumed by the output adapter. The output adapter will 
                // signal that to this thread through a handle.
                EventWaitHandle adapterStopSignal = new EventWaitHandle(false, EventResetMode.ManualReset, "StopAdapter");

                // Start the query and wait for the output adapter to signal 
                // that it has finished receiving events.
                query.Start();

                adapterStopSignal.WaitOne();

                // Retrieve some diagnostic information from the CEP server
                // about the query. Refer to the README.txt of this project on
                // how to use the debugger tool to access diagnostic views
                // remotely.
                tracer.WriteLine(string.Empty);
                RetrieveDiagnostics(server.GetDiagnosticView(new Uri("cep:/Server/EventManager")), tracer);
                RetrieveDiagnostics(server.GetDiagnosticView(new Uri("cep:/Server/PlanManager")), tracer);
                RetrieveDiagnostics(server.GetDiagnosticView(new Uri("cep:/Server/Application/ObjectModelSample/Query/TrafficSensorQuery")), tracer);

                query.Stop();

            }
            catch (Exception e)
            {
                tracer.WriteLine(e.ToString());
            }
            finally
            {
                if (server != null)
                {
                    server.Dispose();
                }
            }

            Console.WriteLine("\npress enter to exit application");
            Console.ReadLine();
        }

        private static QueryBinder BindQuery(InputAdapter inputAdapter, OutputAdapter outputAdapter, QueryTemplate queryTemplate)
        {
            // Create a query binder, wrapping the query template.
            QueryBinder queryBinder = new QueryBinder(queryTemplate);

            // Define the runtime configuration for both input adapters.
            var sensorInputConf = new TextFileInputConfig
            {
                InputFileName = @"..\..\..\Input Data\TrafficSensors.csv",
                Delimiter = ',',
                CTIfrequency = 9,
                CultureName = "en-US",
                InputFieldOrders = new List<string>() { "SensorId", "AverageSpeed", "VehicularCount" }
            };
            var locationInputConf = new TextFileInputConfig
            {
                InputFileName = @"..\..\..\Input Data\TrafficSensorLocations.csv",
                Delimiter = ',',
                CTIfrequency = 100,
                CultureName = "en-US",
                InputFieldOrders = new List<string>() { "SensorId", "LocationId" }
            };

            // Define the runtime configuration for the output adapter.
            // Specify an empty file name, which will just dump the output
            // events on the console. Also pass an event handle name to 
            // synchronize the shutdown.
            var outputConf = new TextFileOutputConfig
            {
                OutputFileName = string.Empty,
                Delimiter = '\t',
                AdapterStopSignal = "StopAdapter",
            };

            // Bind input adapters to query template's input streams,
            // applying runtime configuration.
            // In this example, the given input file for sensor input 
            // contains interval events (each sensor reading has a start 
            // and end time), while the location input is represented by 
            // edge events (for each event, the end time is not known in 
            // advance).
            queryBinder.BindProducer("sensorInput", inputAdapter, sensorInputConf, EventShape.Interval);
            queryBinder.BindProducer("locationInput", inputAdapter, locationInputConf, EventShape.Edge);

            // Bind output adapter to query, applying runtime 
            // configuration.
            queryBinder.AddConsumer<TextFileOutputConfig>("queryresult", outputAdapter, outputConf, EventShape.Point, StreamEventOrder.FullyOrdered);
            return queryBinder;
        }

        private static QueryTemplate CreateQueryTemplate(Application application)
        {
            // Create stream objects as basis for query template
            // specification. The specified names will be used when binding
            // the query template's inputs to event producers.
            CepStream<SensorReading> sensorStream = CepStream<SensorReading>.Create("sensorInput");
            CepStream<LocationData> locationStream = CepStream<LocationData>.Create("locationInput");

            // Extend duration of each sensor reading, so that they fall in
            // a one-minute sliding window. Group by sensor ID and calculate the 
            // average vehicular count per group within each window.
            // Include the grouping key in the aggregation result.
            var avgCount = from oneMinReading in sensorStream.AlterEventDuration(e => TimeSpan.FromMinutes(1))
                           group oneMinReading by oneMinReading.SensorId into oneGroup
                           from eventWindow in oneGroup.SnapshotWindow(SnapshotWindowOutputPolicy.Clip)
                           select new { avgCount = eventWindow.Avg(e => e.VehicularCount), SensorId = oneGroup.Key };

            // Join sensors and locations. Moreover, filter the count 
            // result by a threshold, which is looked up based on the
            // sensor location through a user-defined function.
            var joined = from averageEvent in avgCount
                         join locationData in locationStream
                         on averageEvent.SensorId equals locationData.SensorId
                         where averageEvent.avgCount > UserFunctions.LocationCountThreshold(locationData.LocationId)
                         select new
                         {
                             SensorId = locationData.SensorId,
                             LocationID = locationData.LocationId,
                             VehicularCount = averageEvent.avgCount
                         };

            // BB: Needs description
            return application.CreateQueryTemplate("SampleQueryTemplate", "SampleQueryTemplate Description", joined);
        }

        /// <summary>
        /// Takes a diagnostic view and dumps all its entries to the given 
        /// tracer.
        /// </summary>
        /// <param name="diagview"></param>
        /// <param name="traceListener"></param>
        private static void RetrieveDiagnostics(DiagnosticView diagview, TraceListener traceListener)
        {
            // Display diagnostics for diagnostic view object
            traceListener.WriteLine("Diagnostic View for '" + diagview.ObjectName + "':");
            foreach (KeyValuePair<string, object> diagprop in diagview)
            {
                traceListener.WriteLine(" " + diagprop.Key + ": " + diagprop.Value);
            }
        }
    }
}
