﻿namespace StreamInsight.Samples.ImplicitServer
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Threading;
    using Microsoft.ComplexEventProcessing;
    using Microsoft.ComplexEventProcessing.Linq;
    using StreamInsight.Samples.InputAdapters.TextFile;
    using StreamInsight.Samples.OutputAdapters.TextFile;

    /// <summary>
    /// Console application that demonstrates the implicit server development model.
    /// </summary>
    public class ImplicitServer
    {
        public static void Main(string[] args)
        {
            // Create a tracer to output information on the console.
            TraceListener tracer = new ConsoleTraceListener();

            try
            {
                //BB: Added these lines
                Server server = Server.Create("default");
                Application application = server.CreateApplication("ImplicitServer");

                tracer.WriteLine("Create event producer");
                var inputDataStream = CreateEventProducer();

                // Build the query logic on the input stream
                // BB: Changed form Snapshot
                var top = (from windows in inputDataStream.SnapshotWindow(SnapshotWindowOutputPolicy.Clip)
                           from sampleEvent in windows
                           orderby sampleEvent.Value descending
                           select sampleEvent).Take(1);

                // Create query from the LINQ statement and plug it into the 
                // consumer
                tracer.WriteLine("Create event consumer");
                // BB: Pass app into this method
                Query query = ConsumeQuery(application, top);

                tracer.WriteLine("Start query");

                EventWaitHandle adapterStopSignal = new EventWaitHandle(false, EventResetMode.ManualReset, "StopAdapter");

                // Start the query and wait for the output adapter to signal 
                // that it has finished receiving events.
                query.Start();
                adapterStopSignal.WaitOne();
                query.Stop();
            }
            catch (Exception e)
            {
                tracer.WriteLine(e.ToString());
            }

            Console.WriteLine("\npress enter to exit application");
            Console.ReadLine();
        }

        // BB: pass in and use application
        private static Query ConsumeQuery(Application application, CepStream<SampleReading> stream)
        {
            // Configure text file output output adapter: filename,
            // delimiter, completion event handle name.
            // By using an empty file name, the adapter will dump the 
            // result on the console.
            var outputConf = new TextFileOutputConfig
            {
                OutputFileName = string.Empty,
                Delimiter = '\t',
                AdapterStopSignal = "StopAdapter"
            };

            // Consume the query directly from the output adapter 
            // factory, by telling the factory to instantiate an adapter
            // for interval events
            return stream.ToQuery(application, "SampleQuery", "Description of Sample Query",
                typeof(TextFileOutputFactory), outputConf, EventShape.Interval, StreamEventOrder.FullyOrdered);
        }

        private static CepStream<SampleReading> CreateEventProducer()
        {
            // Configure text file input adapter: filename, delimiter, CTI 
            // frequency
            var inputConfData = new TextFileInputConfig
            {
                InputFileName = @"..\..\..\Input Data\SampleEdgeInput.csv",
                Delimiter = ',',
                CTIfrequency = 1,
                CultureName = "en-US",
                InputFieldOrders = new List<string>() { "SourceId", "Value" }
            };

            // Get a stream directly from the input adapter factory, by 
            // telling the factory to instantiate an adapter with the point
            // event property and our input adapter configuration.
            var inputDataStream = CepStream<SampleReading>.Create("DataInput", typeof(TextFileInputFactory), inputConfData, EventShape.Edge);

            return inputDataStream;
        }
    }
}
