﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.ComplexEventProcessing.Adapters;
using Microsoft.ComplexEventProcessing;
using System.Diagnostics;
using System.Threading;

namespace StreamInsight.Samples.InputAdapters.DataSimulator
{
    public class SimulatedDataEdgeInput<TPayload> : TypedEdgeInputAdapter<TPayload>
    {
        /// <summary>
        /// The configuration information passed into the constructor
        /// </summary>
        private SimulatedDataInputConfig _config;

        /// <summary>
        /// Variable used to swap between start and end edge types when
        /// raising events
        /// </summary>
        private EdgeType _edgeType = EdgeType.Start;

        /// <summary>
        /// Initialize the point input adapter based on the payload type and the 
        /// input configuration.
        /// </summary>
        /// <param name="config"></param>
        public SimulatedDataEdgeInput(SimulatedDataInputConfig config)
        {
            this._config = config;
        }

        public override void Resume()
        {
            ProduceEvents();
        }

        public override void Start()
        {
            ProduceEvents();
        }

        private void ProduceEvents()
        {
            EdgeEvent<TPayload> currEvent =
                default(EdgeEvent<TPayload>);
            EnqueueOperationResult result = EnqueueOperationResult.Full;
            Random rand = new Random();

            // keep state for proper end edge events:
            DateTimeOffset previousStartTime = default(DateTimeOffset);
            TPayload previousPayload = default(TPayload);
            _edgeType = EdgeType.Start;

            while (true)
            {
                if (AdapterState.Stopping == AdapterState)
                {
                    // Signal that we are stopped and exit the method
                    Stopped();
                    return;
                }

                currEvent = CreateInsertEvent(_edgeType);
                if (null == currEvent)
                {
                    continue;
                }

                if (_edgeType == EdgeType.Start)
                {
                    currEvent.StartTime = DateTime.Now;

                    // Using Activator.CreateInstance as we can't add a where TPayload : new() 
                    // constraint (would have to modify the base class to add the constraint)
                    currEvent.Payload = (TPayload)Activator.CreateInstance(typeof(TPayload));

                    if (currEvent.Payload is IRandomInit)
                        (currEvent.Payload as IRandomInit).Init();

                    // remember start tiem and payload for the next end edge
                    previousPayload = currEvent.Payload;
                    previousStartTime = currEvent.StartTime;

                    _edgeType = EdgeType.End;
                }
                else
                {
                    // submit end edge that matches the previous start edge
                    currEvent.StartTime = previousStartTime;
                    currEvent.EndTime = DateTime.Now;
                    currEvent.Payload = previousPayload;

                    _edgeType = EdgeType.Start;
                }

                // Enqueue the event and check for success
                result = Enqueue(ref currEvent);
                if (EnqueueOperationResult.Full == result)
                {
                    ReleaseEvent(ref currEvent);
                    Ready();
                    return;
                }

                Thread.Sleep(_config.EventInterval);
            }
        }
    }
}
