﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.ComplexEventProcessing.Adapters;
using Microsoft.ComplexEventProcessing;

namespace StreamInsight.Samples.InputAdapters.DataSimulator
{
    /// <summary>
    /// This CEP input adapter generates a continuous stream of simulated events, based on:
    /// - Payload type (any type that implements the IRandomInit interface)
    /// - Event settings (how often to raise events)
    /// 
    /// This adapter is intended for testing queries and exploring the capabilities of
    /// StreamInsight.
    /// </summary>
    /// 
    // BB: second interface is new
    public class SimulatedDataInputFactory : ITypedInputAdapterFactory<SimulatedDataInputConfig>,
                                             ITypedDeclareAdvanceTimeProperties<SimulatedDataInputConfig>
    {
        /// <summary>
        /// Create an instance of the specific simulated data input adapter, based:
        /// - Payload type, which must implement IRandomInit
        /// - Configuration information (CTI frequency and interval between events)
        /// - Event shape (point, interval, edge)
        /// </summary>
        /// <typeparam name="TPayload"></typeparam>
        /// <param name="configInfo"></param>
        /// <param name="eventShape"></param>
        /// <returns></returns>
        public InputAdapterBase Create<TPayload>(SimulatedDataInputConfig configInfo, EventShape eventShape)
        {
            InputAdapterBase adapter = default(InputAdapterBase);

            switch (eventShape)
            {
                case EventShape.Edge:
                    adapter = new SimulatedDataEdgeInput<TPayload>(configInfo);
                    break;

                case EventShape.Interval:
                    adapter = new SimulatedDataIntervalInput<TPayload>(configInfo);
                    break;

                case EventShape.Point:
                    adapter = new SimulatedDataPointInput<TPayload>(configInfo);
                    break;
            }

            return adapter;
        }

        public void Dispose()
        {
        }

        /// <summary>
        /// Declare CTI issuing declaratively and factory-wide. Each adapter
        /// instantiated by this factory will automatically insert CTIs as
        /// specified. The frequency is taken from the user-supplied config
        /// info, while the timestamping delay is hard-coded to 0 here.
        /// </summary>
        /// <typeparam name="TPayload"></typeparam>
        /// <param name="configInfo"></param>
        /// <param name="eventShape"></param>
        /// <returns></returns>
        /// BB: new method on interface
        public AdapterAdvanceTimeSettings DeclareAdvanceTimeProperties<TPayload>(SimulatedDataInputConfig configInfo, EventShape eventShape)
        {
            return new AdapterAdvanceTimeSettings(new AdvanceTimeGenerationSettings(
                configInfo.CtiFrequency, TimeSpan.FromSeconds(0)), AdvanceTimePolicy.Drop);
        }

    }
}
