﻿namespace StreamInsight.Samples.InputAdapters.Sql
{
    using System;
    using System.Data.SqlClient;
    using StreamInsight.Samples.EventTypes;
    using Microsoft.ComplexEventProcessing;
    using Microsoft.ComplexEventProcessing.Adapters;

    /// <summary>
    /// This typed adapter in this sample populates the event payload 
    /// structure from specific fields projected by a SQL query (shown below in
    /// this sample code) against the table
    ///   [AdventureWorks2008].[Sales].[SalesOrderDetail]
    /// from the AdventureWorks2008 sample database that is downloadable from 
    /// codeplex.com (Bing search for "SQL Server 2008 Sample Database" to find
    /// the exact site, and download). The example uses DateTimeOffset.UtcNow (current
    /// system time) as the StartTime of the event.
    /// </summary>
    public class SqlTypedPointInput : TypedPointInputAdapter<SalesOrderDetail>
    {
        private const int Numberofticks = 0;
        private const int Zero = 0;

        private SqlConnection connection = null;
        private SqlDataReader reader = null;
        private int ctiFrequency;
        private int _eventsEnqueued = 0;
        private PointEvent<SalesOrderDetail> pendingEvent;
        private DateTimeOffset? pendingCtiTime = null;

        public SqlTypedPointInput(SqlInputConfig configinfo)
        {
            string sqlstmt = @"select [SalesOrderID], [SalesOrderDetailID], [OrderQty], Convert(float, [UnitPrice]) AS UnitPrice from [Sales].[SalesOrderDetail]";
            ctiFrequency = configinfo.CtiFrequency;

            try
            {
                connection = new SqlConnection(configinfo.ConnectionString);
                connection.Open();

                SqlCommand command = new SqlCommand(sqlstmt, connection);
                reader = command.ExecuteReader();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message + e.StackTrace);
            }
        }

        public override void Start()
        {
            ProduceEvents();
        }

        public override void Resume()
        {
            ProduceEvents();
        }

        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);
        }

        private void ProduceEvents()
        {
            PointEvent<SalesOrderDetail> currEvent = default(PointEvent<SalesOrderDetail>);
            EnqueueOperationResult result = EnqueueOperationResult.Full;
            bool hasRow = false;

            try
            {
                while (true)
                {
                    if (AdapterState.Stopping == AdapterState)
                    {
                        // first, check if the adapter is being asked to stop. If yes, then
                        // resolve the event from the last failed Enqueue;
                        // do housekeeping before calling stopped;
                        // declare the adapter as stopped; and then
                        // exit the worker thread
                        // Note: Given that it is the engine that is asking the adapter to
                        // stop, we do not enqueue a Cti with a max time value here - this
                        // will cause an exception because the engine is not expecting any
                        // more inputs from the adapter.
                        if (pendingEvent != null)
                        {
                            currEvent = pendingEvent;
                            pendingEvent = null;
                        }

                        PrepareToStop(currEvent);
                        Stopped();
                        return;
                    }

                    // NOTE: at any point during execution of the code below, if the Adapter state
                    // changes to Stopping, the engine will resume the adapter (i.e. call Resume())
                    // one more time, and the stopping condition will be trapped at the check above.
                    //
                    // If a previous enqueue failed, enqueue the pending event first before creating
                    // a new event
                    // check for a pending point event, if none,
                    // check for a pending Cti event, if none,
                    // create an event from the file
                    if (pendingEvent != null)
                    {
                        currEvent = pendingEvent;
                        pendingEvent = null;
                    }
                    else if (pendingCtiTime != null)
                    {
                        // the check is important; but for this adapter, there is nothing to be done here.
                        // In general, you may want to take some action if the last CTI dequeue failed
                    }
                    else
                    {
                        hasRow = reader.Read();
                        if (hasRow)
                        {
                            // Create Event from row
                            currEvent = CreateInsertEvent();

                            // DateTime.Now is the current system time. Its value in the next iteration
                            // will be a few milliseconds away from the current reading.
                            currEvent.StartTime = DateTimeOffset.UtcNow;

                            // fill the payload fields from the table
                            currEvent.Payload = new SalesOrderDetail(
                                reader.GetInt32(reader.GetOrdinal("SalesOrderID")),
                                reader.GetInt32(reader.GetOrdinal("SalesOrderDetailID")),
                                reader.GetInt16(reader.GetOrdinal("OrderQty")),
                                reader.GetDouble(reader.GetOrdinal("UnitPrice")));
                            pendingEvent = null;
                        }
                        else
                        {
                            // if end of table, issue a max Cti (to indicate that there will be no
                            // more event arrivals with timestamp earlier than this Cti)
                            // prepare to stop, indicate Stopped and exit from worker thread
                            // EnqueueCtiEvent(DateTime.MaxValue);
                            PrepareToStop(currEvent);
                            Stopped();
                            return;
                        }
                    }

                    // Enqueue the Cti event (or) Current event with payload.
                    // If Enqueue is rejected, save the Cti time or the event itself in
                    // PrepareToResume, indicate Ready to resume, and exit worker thread
                    if (pendingCtiTime != null)
                    {
                        result = EnqueueCtiEvent(pendingCtiTime.Value);
                        if (EnqueueOperationResult.Full == result)
                        {
                            PrepareToResume(pendingCtiTime.Value);
                            Ready();
                            return;
                        }
                        else
                        {
                            // there is no Cti event pending
                            pendingCtiTime = null;
                        }
                    }
                    else
                    {
                        // In this example, the StartTime of the Cti event is a few ticks past the
                        // StartTime of the previously enqueued point event. However, Enqueue releases
                        // the memory allocated for the event - so currEvent will not be available in the
                        // EnqueueCti code block below. So in anticipation of a successful enqueue,
                        // save away the StartTime of the event for enqueueing Cti events.
                        DateTimeOffset currEventTime = currEvent.StartTime;

                        // enqueue event with payload
                        PrintEvent(currEvent);
                        result = Enqueue(ref currEvent);

                        if (EnqueueOperationResult.Full == result)
                        {
                            PrepareToResume(currEvent);
                            Ready();
                            return;
                        }
                        else
                        {
                            // if the enqueue was successful, set no pending events, and check for CTI injection
                            _eventsEnqueued++;
                            pendingEvent = null;

                            // Enqueue CTI according to prescribed frequency, setting the Cti time as
                            // number of ticks past the application timestamp of the last enqueued point event
                            if (Zero == (_eventsEnqueued % ctiFrequency))
                            {
                                DateTimeOffset currCtiTime = currEventTime + TimeSpan.FromTicks(Numberofticks);

                                Console.WriteLine("Input: CTI|" + currCtiTime + "|");
                                result = EnqueueCtiEvent(currCtiTime);
                                if (EnqueueOperationResult.Full == result)
                                {
                                    PrepareToResume(currCtiTime);
                                    Ready();
                                    return;
                                }
                                else
                                {
                                    // reset the Cti time so that new events can be enqueued
                                    pendingCtiTime = null;
                                }
                            }
                        }
                    }
                }
            }
            catch (AdapterException e)
            {
                Console.WriteLine("ProduceEvents - " + e.Message + e.StackTrace);
            }
        }

        private void PrepareToStop(PointEvent<SalesOrderDetail> currEvent)
        {
            // the server will not accept any more events, and you
            // cannot do anything about it. So release the event.
            // If you miss this step, server memory will leak.
            if (null != currEvent)
            {
                ReleaseEvent(ref currEvent);
            }

            reader.Close();
            connection.Close();
        }

        private void PrepareToResume(PointEvent<SalesOrderDetail> currEvent)
        {
            pendingEvent = currEvent;
        }

        private void PrepareToResume(DateTimeOffset currCtiTime)
        {
            pendingCtiTime = currCtiTime;
        }

        private void PrintEvent(PointEvent<SalesOrderDetail> evt)
        {
            Console.WriteLine("Input: " + evt.EventKind + "|" + evt.StartTime + "|" +
                evt.Payload.SalesOrderId + "|" + evt.Payload.SalesOrderDetailId + "|" +
                evt.Payload.OrderQuantity + "|" + evt.Payload.UnitPrice + "|");
        }
    }
}
