﻿namespace StreamInsight.Samples.InputAdapters.TextFile
{
    using System;
    using System.Collections.Generic;
    using System.Globalization;
    using System.IO;
    using Microsoft.ComplexEventProcessing;
    using Microsoft.ComplexEventProcessing.Adapters;
    
    /// <summary>
    /// TextFileEdgeOutput adapter enqueues generic point events from a CSV file with the format:
    ///            Start{delimiter}{StartTime}{delimiter}{delimiter}{field1}{delimiter}...{fieldN}
    ///            End{delimiter}{StartTime}{delimiter}{Endtime}{delimiter}{field1}{delimiter}...{fieldN}
    /// The specific file that works with this example is: TextFileEdgeInput.csv
    /// Edge Events are used to model interval events whose EndTime of the interval is not known
    /// at the time of the start of the event. Rather than wait for end of the interval at a cost
    /// to the overall latency of the system, you can specify the Start edge with just the Start time
    /// and the payload values, to be subsequently followed by the end edge. A few notes:
    /// - the EndTime for the Start Edge is a don't care - the engine treats the value as +Infinity.
    ///   In this example, the EndTime for Start Edge is an empty string (between the two successive
    ///   delimiters shown above). This requires some accommodation in the use of split function while
    ///   reading the file input.
    /// - the End Edge should be specified to have the same StartTime and payload fields as the Start Edge.
    /// - every start edge should have a matching end edge
    /// - Cti's can be issued between a start edge and end edge. However, you cannot specify an end edge
    ///   with its EndTimestamp earlier than that of the Cti's timestamp. The engine will raise an exception
    ///   upon such input.
    /// <para/>
    /// See additional notes in TextFilePointInput.cs
    /// </summary>
    public class TextFileEdgeInput : EdgeInputAdapter
    {
        private const int Zero = 0;
        private const int NumberOfTicks = 0;
        private const int NumNonPayloadFields = 3;

        private StreamReader _streamReader;
        private string _line;
        private char _delimiter;
        private CultureInfo _culture;
        private Dictionary<int, int> _inputOrdinalToCepOrdinal;

        private CepEventType _bindtimeEventType;
        private int _ctiFrequency;
        private int _eventsEnqueued;

        private EdgeEvent _pendingEvent;
        private DateTimeOffset? _pendingCtiTime = null;

        /// <summary>
        /// Initializes a new instance of the TextFileEdgeInput class.
        /// See notes on the Constructor Method in TextFilePointInput.cs
        /// </summary>
        /// <param name="configInfo"></param>
        /// <param name="cepEventType"></param>
        public TextFileEdgeInput(TextFileInputConfig configInfo, CepEventType cepEventType)
        {
            _streamReader = new StreamReader(configInfo.InputFileName);
            _ctiFrequency = configInfo.CTIfrequency;
            _bindtimeEventType = cepEventType;
            _delimiter = configInfo.Delimiter;
            _culture = new CultureInfo(configInfo.CultureName);

            // Create mapping from input field ordinal to engine field ordinal
            _inputOrdinalToCepOrdinal = new Dictionary<int, int>();
            
            if (configInfo.InputFieldOrders != null && configInfo.InputFieldOrders.Count != cepEventType.Fields.Count)
            {
                throw new ArgumentException(
                    string.Format(CultureInfo.InvariantCulture,
                    "The configuration element InputFieldOrders should have {0} elements, but it only has {1} elements", 
                    cepEventType.Fields.Count,
                    configInfo.InputFieldOrders.Count));
            }

            CepEventTypeField engineField;
            for (int i = 0; i < cepEventType.Fields.Count; i++)
            {
                if (configInfo.InputFieldOrders != null)
                {
                    if (!cepEventType.Fields.TryGetValue(configInfo.InputFieldOrders[i], out engineField))
                    {
                        throw new ArgumentException(
                            string.Format(CultureInfo.InvariantCulture,
                            "Event type {0} doesn't have an input field named '{1}'", cepEventType.ShortName, configInfo.InputFieldOrders[i]));
                    }

                    _inputOrdinalToCepOrdinal.Add(i, engineField.Ordinal);
                }
                else
                {
                    // Use default mapping
                    _inputOrdinalToCepOrdinal.Add(i, i);
                }
            }
        }

        /// <summary>
        /// See notes on the Start() method in TextFilePointInput.cs
        /// </summary>
        public override void Start()
        {
            ProduceEvents();
        }

        /// <summary>
        /// See notes on the Resume() method in TextFilePointInput.cs
        /// </summary>
        public override void Resume()
        {
            ProduceEvents();
        }

        /// <summary>
        /// Default Dispose from the base class.
        /// </summary>
        /// <param name="disposing"></param>
        protected override void Dispose(bool disposing)
        {
            _streamReader.Dispose();
            base.Dispose(disposing);
        }

        /// <summary>
        /// Main driver to read events from the CSV file and enqueue them
        /// </summary>
        private void ProduceEvents()
        {
            EdgeEvent currEvent = default(EdgeEvent);
            EnqueueOperationResult result = EnqueueOperationResult.Full;

            try
            {
                while (true)
                {
                    if (AdapterState.Stopping == AdapterState)
                    {
                        // first, check if the adapter is being asked to stop. 
                        // If yes, then resolve the event from the last failed
                        // Enqueue; do housekeeping before calling stopped;
                        // declare the adapter as stopped; and then
                        // exit the worker thread
                        if (_pendingEvent != null)
                        {
                            currEvent = _pendingEvent;
                            _pendingEvent = null;
                        }

                        PrepareToStop(currEvent);
                        Stopped();
                        return;
                    }

                    // NOTE: at any point during execution of the code below, 
                    // if the Adapter state changes to Stopping, the engine
                    // will resume the adapter (i.e. call Resume()) one more
                    // time, and the stopping condition will be trapped at the 
                    // check above.

                    // if a previous enqueue failed, enqueue the pending event
                    // first before creating a new event;
                    // check for a pending payload event, if none,
                    // check for a pending Cti event, if none,
                    // create an event from the file
                    if (_pendingEvent != null)
                    {
                        currEvent = _pendingEvent;
                        _pendingEvent = null;
                    }
                    else if (_pendingCtiTime != null)
                    {
                        // the check is important; but for this adapter, there 
                        // is nothing to be done here. In general, you may want
                        // to take some action if the last CTI enqueue failed
                    }
                    else
                    {
                        _line = _streamReader.ReadLine();
                        if (null != _line)
                        {
                            currEvent = CreateEventFromLine(_line);
                            _pendingEvent = null;
                        }
                        else
                        {
                            // if end of file, prepare to stop, indicate Stopped and exit from worker thread
                            EnqueueCtiEvent(DateTimeOffset.MaxValue);
                            PrepareToStop(currEvent);
                            Stopped();
                            return;
                        }
                    }

                    // Enqueue the Cti event (or) Current event with payload.
                    // If Enqueue is rejected, save the Cti time or the event itself in
                    // PrepareToResume, indicate Ready to resume, and exit worker thread
                    if (_pendingCtiTime != null)
                    {
                        result = EnqueueCtiEvent(_pendingCtiTime.Value);
                        if (EnqueueOperationResult.Full == result)
                        {
                            PrepareToResume(_pendingCtiTime.Value);
                            Ready();
                            return;
                        }
                        else
                        {
                            // there is no Cti event pending
                            _pendingCtiTime = null;
                        }
                    }
                    else
                    {
                        // In this example, the StartTime of the Cti event is a few (>= 1) ticks past the
                        // EndTime of the previously enqueued Edge event. However, Enqueue releases
                        // the memory allocated for the event - so currEvent will not be available in the
                        // EnqueueCti code block below. So in anticipation of a successful enqueue,
                        // save away the Edge time of the event for enqueueing Cti events.
                        DateTimeOffset currEdgeTime =
                            (EdgeType.Start == currEvent.EdgeType) ? currEvent.StartTime : currEvent.EndTime;
                        
                        // enqueue event with payload
                        result = Enqueue(ref currEvent);

                        if (EnqueueOperationResult.Full == result)
                        {
                            PrepareToResume(currEvent);
                            Ready();
                            return;
                        }
                        else
                        {
                            // if the enqueue was successful, set no pending events,
                            // and check for CTI injection
                            _eventsEnqueued++;
                            _pendingEvent = null;

                            // Enqueue CTI according to prescribed frequency, setting the Cti time as
                            // number of ticks past the end timestamp of the last enqueued edge event
                            if (Zero == (_eventsEnqueued % _ctiFrequency))
                            {
                                DateTimeOffset currCtiTime = currEdgeTime + TimeSpan.FromTicks(NumberOfTicks);
                                result = EnqueueCtiEvent(currCtiTime);
                                if (EnqueueOperationResult.Full == result)
                                {
                                    PrepareToResume(currCtiTime);
                                    Ready();
                                    return;
                                }
                                else
                                {
                                    // reset the Cti time so that new events can be enqueued
                                    _pendingCtiTime = null;
                                }
                            }
                        }
                    }
                }
            }
            catch (AdapterException e)
            {
                Console.WriteLine("ProduceEvents - " + e.Message + e.StackTrace);
            }
        }

        private void PrepareToStop(EdgeEvent currEvent)
        {
            // the server will not accept any more events, and you
            // cannot do anything about it. So release the event.
            // If you miss this step, server memory will leak.
            if (null != currEvent)
            {
                ReleaseEvent(ref currEvent);
            }
        }

        private void PrepareToResume(EdgeEvent currEvent)
        {
            _pendingEvent = currEvent;
        }

        private void PrepareToResume(DateTimeOffset currCtiTime)
        {
            _pendingCtiTime = currCtiTime;
        }

        private EdgeEvent CreateEventFromLine(string line)
        {
            // use a split that does NOT remove empty entries, so that the payload load indexing
            // remains the same as for interval events.
            string[] split = line.Split(new char[] { _delimiter }, StringSplitOptions.None);

            EdgeEvent evt = CreateInsertEvent(("Start" == split[0]) ? EdgeType.Start : EdgeType.End);
            evt.StartTime = DateTime.Parse(split[1], _culture, DateTimeStyles.AssumeUniversal).ToUniversalTime();
            if (evt.EdgeType == EdgeType.End)
            {
                evt.EndTime = DateTime.Parse(split[2], _culture, DateTimeStyles.AssumeUniversal).ToUniversalTime();
            }

            // populate the payload fields
            for (int ordinal = 0; ordinal < _bindtimeEventType.FieldsByOrdinal.Count; ordinal++)
            {
                try
                {
                    int cepOrdinal = _inputOrdinalToCepOrdinal[ordinal];
                    CepEventTypeField evtField = _bindtimeEventType.FieldsByOrdinal[cepOrdinal];

                    object value = Convert.ChangeType(split[ordinal + NumNonPayloadFields], evtField.Type.ClrType, _culture);
                    evt.SetField(cepOrdinal, value);
                }
                catch (AdapterException e)
                {
                    Console.WriteLine(e.Message + e.StackTrace);
                }
            }

            return evt;
        }
    }
}