﻿namespace StreamInsight.Samples.InputAdapters.TextFile
{
    using System;
    using System.Collections.Generic;
    using System.Globalization;
    using System.IO;
    using Microsoft.ComplexEventProcessing;
    using Microsoft.ComplexEventProcessing.Adapters;

    /// <summary>
    /// TextFilePointOutput adapter enqueues generic point events from a CSV file with the format:
    ///   {StartTime}{delimiter}{field1}{delimiter}...{fieldN}
    /// The specific file that works with this example is: SamplePointInput.csv
    /// <para/>
    /// By generic events, we mean that, for a given input device (in this case a CSV file), you
    /// can specify an arbitrary payload class or struct at the time of query binding. This enables
    /// you to reuse this adapter implementation for multiple instantiations of the adapter - where
    /// each instantiation corresponds to a particular payload type, and is associated/bound 1:1
    /// with a query.
    /// <para/>
    /// Please study this code in conjunction with the adapter state transition diagram provided in
    /// the documentation.
    /// </summary>
    public class TextFilePointInput : PointInputAdapter
    {
        private const int Zero = 0;
        private const int NumberOfTicks = 3;
        private const int NumNonPayloadFields = 1;

        private StreamReader _streamReader;
        private string _line;
        private char _delimiter;
        private CultureInfo _culture;
        private Dictionary<int, int> _inputOrdinalToCepOrdinal;

        private CepEventType _bindtimeEventType;
        private int _ctiFrequency;
        private int _eventsEnqueued;

        private PointEvent _pendingEvent;
        private DateTimeOffset? _pendingCtiTime = null;
        
        private System.Diagnostics.TraceListener _consoleTracer;

        /// <summary>
        /// Initializes a new instance of the TextFilePointInput class.
        /// <para/>
        /// Notes on the Constructor Method:
        /// TextFileInputConfig is provided via the TextFileInputFactory class, and contains
        /// the following configuration information:
        ///  * Filename
        ///  * Delimiter
        ///  * CTI Frequency
        /// The filename is a parameter to the StreamReader method - which 
        /// will retrieve each line as a string using synchronous reads. This
        /// string is parsed and split into individual fields along the
        /// specified delimiter character. The type of each field is inferred 
        /// using a CLR method. The event is then composed on a field by field
        /// basis. Each field type maps to a scalar, CLR type. The CTI
        /// Frequency specified in the configuration structure determines 
        /// after how many events a CTI is to be enqueued.
        /// <para/>
        /// CepEventType represents the actual event type that will be passed into this
        /// constructor from the query binding via the adapter factory class. The payload class
        /// or struct supports several CLR types. This example can handle most scalar fields.
        /// </summary>
        /// <param name="configInfo"></param>
        /// <param name="cepEventType"></param>
        public TextFilePointInput(TextFileInputConfig configInfo, CepEventType cepEventType)
        {
            _streamReader = new StreamReader(configInfo.InputFileName);
            _ctiFrequency = configInfo.CTIfrequency;
            _bindtimeEventType = cepEventType;
            _delimiter = configInfo.Delimiter;
            _culture = new CultureInfo(configInfo.CultureName);

            // Create mapping from input field ordinal to engine field ordinal
            _inputOrdinalToCepOrdinal = new Dictionary<int, int>();

            if (configInfo.InputFieldOrders != null && configInfo.InputFieldOrders.Count != cepEventType.Fields.Count)
            {
                throw new ArgumentException(
                    string.Format(CultureInfo.InvariantCulture,
                    "The configuration element InputFieldOrders should have {0} elements, but it only has {1} elements",
                    cepEventType.Fields.Count,
                    configInfo.InputFieldOrders.Count));
            }

            CepEventTypeField engineField;
            for (int i = 0; i < cepEventType.Fields.Count; i++)
            {
                if (configInfo.InputFieldOrders != null)
                {
                    if (!cepEventType.Fields.TryGetValue(configInfo.InputFieldOrders[i], out engineField))
                    {
                        throw new ArgumentException(
                            string.Format(CultureInfo.InvariantCulture,
                            "Event type {0} doesn't have an input field named '{1}'", cepEventType.ShortName, configInfo.InputFieldOrders[i]));
                    }

                    _inputOrdinalToCepOrdinal.Add(i, engineField.Ordinal);
                }
                else
                {
                    // Use default mapping
                    _inputOrdinalToCepOrdinal.Add(i, i);
                }
            }

            _consoleTracer = new System.Diagnostics.ConsoleTraceListener();
        }

        /// <summary>
        /// Start is the first method to be invoked by the CEP engine once the input adapter has been instantiated.
        /// So any initializations that cannot be covered in the constructor can be placed here. Start() is called
        /// in a separate worker thread initiated by the engine. In this example (and in most scenarios), Start
        /// begins producing events once it is invoked.
        /// </summary>
        public override void Start()
        {
            ProduceEvents();
        }

        /// <summary>
        /// Resume is invoked by the engine once it returns from being scheduled away from Start, and only after
        /// the adapter has called Ready(). Resume continues to produce events.
        /// </summary>
        public override void Resume()
        {
            ProduceEvents();
        }

        /// <summary>
        /// Default Dispose from the base class.
        /// </summary>
        /// <param name="disposing"></param>
        protected override void Dispose(bool disposing)
        {
            _streamReader.Dispose();
            base.Dispose(disposing);
        }

        /// <summary>
        /// Main driver to read events from the CSV file and enqueue them
        /// </summary>
        private void ProduceEvents()
        {
            PointEvent currEvent = default(PointEvent);
            EnqueueOperationResult result = EnqueueOperationResult.Full;

            try
            {
                while (true)
                {
                    if (AdapterState.Stopping == AdapterState)
                    {
                        // first, check if the adapter is being asked to stop. If yes, then
                        // resolve the event from the last failed Enqueue;
                        // do housekeeping before calling stopped;
                        // declare the adapter as stopped; and then
                        // exit the worker thread
                        if (_pendingEvent != null)
                        {
                            currEvent = _pendingEvent;
                            _pendingEvent = null;
                        }

                        PrepareToStop(currEvent);
                        Stopped();
                        return;
                    }

                    // NOTE: at any point during execution of the code below, if the Adapter state
                    // changes to Stopping, the engine will resume the adapter (i.e. call Resume())
                    // one more time, and the stopping condition will be trapped at the check above.
                    
                    // if a previous enqueue failed, enqueue the pending event first before creating a new event
                    // check for a pending point event, if none,
                    // check for a pending Cti event, if none,
                    // create an event from the file
                    if (_pendingEvent != null)
                    {
                        currEvent = _pendingEvent;
                        _pendingEvent = null;
                    }
                    else if (_pendingCtiTime != null)
                    {
                        // the check is important; but for this adapter, there is nothing to be done here.
                        // In general, you may want to take some action if the last CTI dequeue failed
                    }
                    else
                    {
                        _line = _streamReader.ReadLine();
                        if (null != _line)
                        {
                            try
                            {
                                currEvent = CreateEventFromLine(_line);
                            }
                            catch (Exception e)
                            {
                                _consoleTracer.WriteLine(_line + "could not be read into a CEP event: " + e.Message);
                                continue;
                            }

                            _pendingEvent = null;
                        }
                        else
                        {
                            // if end of file, issue a max Cti (to indicate that there will be no
                            // more event arrivals with timestamp earlier than this Cti)
                            // prepare to stop, indicate Stopped and exit from worker thread
                            EnqueueCtiEvent(DateTimeOffset.MaxValue);
                            PrepareToStop(currEvent);
                            Stopped();
                            return;
                        }
                    }

                    // Enqueue the Cti event (or) Current event with payload.
                    // If Enqueue is rejected, save the Cti time or the event itself in
                    // PrepareToResume, indicate Ready to resume, and exit worker thread
                    if (_pendingCtiTime != null)
                    {
                        result = EnqueueCtiEvent(_pendingCtiTime.Value);
                        if (EnqueueOperationResult.Full == result)
                        {
                            PrepareToResume(_pendingCtiTime.Value);
                            Ready();
                            return;
                        }
                        else
                        {
                            // there is no Cti event pending
                            _pendingCtiTime = null;
                        }
                    }
                    else
                    {
                        // In this example, the StartTime of the Cti event is a few (>= 2) ticks past the
                        // StartTime of the previously enqueued point event. However, Enqueue releases
                        // the memory allocated for the event - so currEvent will not be available in the
                        // EnqueueCti code block below. So in anticipation of a successful enqueue,
                        // save away the StartTime of the event for enqueueing Cti events.
                        DateTimeOffset currEventTime = currEvent.StartTime;

                        // enqueue event with payload
                        result = Enqueue(ref currEvent);

                        if (EnqueueOperationResult.Full == result)
                        {
                            PrepareToResume(currEvent);
                            Ready();
                            return;
                        }
                        else
                        {
                            // if the enqueue was successful, set no pending events, and check for CTI injection
                            _eventsEnqueued++;
                            _pendingEvent = null;
                            
                            // Enqueue CTI according to prescribed frequency, setting the Cti time as
                            // number of ticks past the application timestamp of the last enqueued point event
                            if (Zero == (_eventsEnqueued % _ctiFrequency))
                            {
                                DateTimeOffset currCtiTime = currEventTime + TimeSpan.FromTicks(NumberOfTicks);
                                result = EnqueueCtiEvent(currCtiTime);
                                if (EnqueueOperationResult.Full == result)
                                {
                                    PrepareToResume(currCtiTime);
                                    Ready();
                                    return;
                                }
                                else
                                {
                                    // reset the Cti time so that new events can be enqueued
                                    _pendingCtiTime = null;
                                }
                            }
                        }
                    }
                }
            }
            catch (AdapterException e)
            {
                Console.WriteLine("ProduceEvents - " + e.Message + e.StackTrace);
            }
        }

        private void PrepareToStop(PointEvent currEvent)
        {
            // the server will not accept any more events, and you
            // cannot do anything about it. So release the event.
            // If you miss this step, server memory will leak.
            if (null != currEvent)
            {
                ReleaseEvent(ref currEvent);
            }
        }

        private void PrepareToResume(PointEvent currEvent)
        {
            _pendingEvent = currEvent;
        }

        private void PrepareToResume(DateTimeOffset currCtiTime)
        {
            _pendingCtiTime = currCtiTime;
        }

        // Create event from line of format:
        // Line format: <StartTime>,<comma separated Payload Fields>
        private PointEvent CreateEventFromLine(string line)
        {
            string[] split = line.Split(new char[] { _delimiter }, StringSplitOptions.RemoveEmptyEntries);
            if (_bindtimeEventType.Fields.Count != (split.Length - NumNonPayloadFields))
            {
                throw new Exception("Number of payload columns in input file: " + (split.Length - NumNonPayloadFields) +
                    ", does not match number of fields in EventType: " + _bindtimeEventType.Fields.Count);
            }

            PointEvent evt = CreateInsertEvent();

            // set Start time
            evt.StartTime = DateTime.Parse(split[0], _culture, DateTimeStyles.AssumeUniversal).ToUniversalTime();

            // populate the payload fields
            for (int ordinal = 0; ordinal < _bindtimeEventType.FieldsByOrdinal.Count; ordinal++)
            {
                try
                {
                    int cepOrdinal = _inputOrdinalToCepOrdinal[ordinal];
                    CepEventTypeField evtField = _bindtimeEventType.FieldsByOrdinal[cepOrdinal];

                    object value = Convert.ChangeType(split[ordinal + NumNonPayloadFields], evtField.Type.ClrType, _culture);
                    evt.SetField(cepOrdinal, value);
                }
                catch (AdapterException e)
                {
                    Console.WriteLine(e.Message + e.StackTrace);
                }
            }

            return evt;
        }
    }
}