﻿namespace StreamInsight.Samples.Observable
{
    using System;
    using System.Diagnostics;
    using System.Linq;
    using System.Threading;
    using Microsoft.ComplexEventProcessing.Adapters.Observable;
    using Microsoft.ComplexEventProcessing.Linq;

    /// <summary>
    /// Console Application that demonstrates the Observable programming model.
    /// </summary>
    public class Observable
    {
        public static void Main(string[] args)
        {
            // Create a tracer to output information on the console.
            TraceListener tracer = new ConsoleTraceListener();

            tracer.WriteLine("Setting up event log reader.");

            // Create a Windows event log reader as an IEnumerable
            var eventLogEnum = new EventLog("System", ".").Entries.OfType<EventLogEntry>();

            // Preprocess the event log: just consider the last 30 days worth 
            // of events
            var enumInput = from eventLogEntry in eventLogEnum
                            where eventLogEntry.TimeWritten > DateTime.Now.Subtract(TimeSpan.FromDays(30))
                            select eventLogEntry;

            // Create an observable out of the IEnumerable
            var observableInput = enumInput.ToObservable();

            // Trim strings to a maxlength to keep the event small
            int maxStringLength = 250;

            // The Observable acts as a data source. We can access it as a
            // CepStream to specify the query template on top of it.
            // We specify the entire type mapping, including the mapping
            // from the type's property TimeGenerated to the CEP timestamp.
            var input = observableInput.ToCepStream(e =>
                            new CepPointEvent<MyEventLogEntry>
                            {
                                EventKind = Microsoft.ComplexEventProcessing.EventKind.Insert,
                                StartTime = e.TimeGenerated,
                                Payload = new MyEventLogEntry
                                {
                                    Category = e.Category.Substring(0, Math.Min(maxStringLength, e.Category.Length)),
                                    CategoryNumber = e.CategoryNumber,
                                    Data = e.Data,
                                    EntryType = (int)e.EntryType,
                                    Index = e.Index,
                                    InstanceId = e.InstanceId,
                                    MachineName = null == e.MachineName ? null : e.MachineName.Substring(0, Math.Min(maxStringLength, e.MachineName.Length)),
                                    Message = null == e.Message ? null : e.Message.Substring(0, Math.Min(maxStringLength, e.Message.Length)),
                                    Source = null == e.Source ? null : e.Source.Substring(0, Math.Min(maxStringLength, e.Source.Length)),
                                    TimeGenerated = e.TimeGenerated,
                                    TimeWritten = e.TimeWritten,
                                    UserName = null == e.UserName ? null : e.UserName.Substring(0, Math.Min(maxStringLength, e.UserName.Length)),
                                }
                            });

            var queryOutputStream = CreateQueryTemplate(input);

            // Convert the query template to an Observable, and specify that 
            // the event's timestamp should be mapped into the payload's Time
            // field. By calling ToObservable, we are leaving the CEP paradigm
            // and hence need to carry the CEP timestamps into the CLR payload 
            // type. Alternatively, we could just drop the CEP timestamps, but
            // then we miss that temporal information in the query result.
            var queryOutput = queryOutputStream.ToObservable(
                    (CepPointEvent<EventTypeCount> e) =>
                        new EventTypeCount
                        {
                            EventCount = e.Payload.EventCount,
                            EventType = e.Payload.EventType,
                            Time = e.StartTime.UtcDateTime
                        } );

            ManualResetEvent observerStopSignal = new ManualResetEvent(false);

            // Set up the Observer, which will write the result events as rows 
            // into a file. Also pass the event handle, so that we can wait for the
            // observer to finish.
            var myTextFileWriterObserver = new MyTextFileWriterObserver("DailyEventCounter.csv", observerStopSignal);

            // Add the observer as subscriber to the query. This will also 
            // start the query.
            IDisposable subscription = queryOutput.Subscribe(myTextFileWriterObserver);

            tracer.WriteLine("Waiting for observer to signal completion (this can take a couple of seconds)...");

            observerStopSignal.WaitOne();
        }

        private static CepStream<EventTypeCount> CreateQueryTemplate(CepStream<MyEventLogEntry> input)
        {
            // Define the query logic:
            // Aggregate: Count events for each entry type within the window
            var countPerGroup = from dayEvents in input
                                group dayEvents by dayEvents.EntryType into eachGroup
                                from window in eachGroup.TumblingWindow(TimeSpan.FromDays(1))
                                select new { Count = window.Count(), EntryType = eachGroup.Key };

            // Do a final project to map the entry types to meaningful strings.
            // Project the time field with a dummy value (DateTime.Now) - it will be
            // overwritten below.
            var queryOutputStream = from countEvents in countPerGroup
                                    select new EventTypeCount
                                    {
                                        EventCount = countEvents.Count,
                                        EventType = countEvents.EntryType == (int)EventLogEntryType.Error ? "Error" :
                                                     countEvents.EntryType == (int)EventLogEntryType.Warning ? "Warning" :
                                                      countEvents.EntryType == (int)EventLogEntryType.Information ? "Information" :
                                                       countEvents.EntryType == (int)EventLogEntryType.SuccessAudit ? "SuccessAudit" :
                                                        "FailureAudit",
                                        Time = DateTime.Now
                                    };

            return queryOutputStream;
        }
    }
}
