﻿namespace StreamInsight.Samples.OutputAdapters.Sql
{
    using System;
    using System.Data;
    using System.Data.SqlClient;
    using System.Threading;
    using StreamInsight.Samples.EventTypes;
    using Microsoft.ComplexEventProcessing;
    using Microsoft.ComplexEventProcessing.Adapters;

    /// <summary>
    /// This typed adapter in this sample populates (i.e. INSERTs rows into) a table in the SQL Server
    /// database with the schema:
    ///       (SalesOrderId Integer, TotalSales Money)
    /// The main focus of this sample is in demonstrating how to dequeue events from the CEP Server
    /// and populating the table. So the sample makes the following basic assumptions.
    /// 1. The [database].[schema].[table] specified by you in the configinfo already exists.
    /// 2. You will delete or maintain the results between successive runs of the sample. In this
    ///    example, the query/ app will insert duplicate rows in successive runs.
    /// </summary>
    public class SqlTypedPointOutput : TypedPointOutputAdapter<SalesByOrderId>
    {
        private string insert;
        private SqlConnection connection = null;
        private SqlCommand command = null;
        private EventWaitHandle _adapterStopSignal;
        private int _eventsDequeued = 0;

        public SqlTypedPointOutput(SqlOutputConfig configinfo)
        {
            insert = @"insert into [AdventureWorks2008].[guest].[SalesByOrderId] (EventKind, StartTime, SalesOrderId, TotalSales) values (@EventKind, @StartTime,@SalesOrderId, @TotalSales)";

            try
            {
                connection = new SqlConnection(configinfo.ConnectionString);
                connection.Open();

                command = new SqlCommand(insert, connection);
                command.Parameters.Add("@EventKind", SqlDbType.NChar, 10, "EventKind");
                command.Parameters.Add("@StartTime", SqlDbType.DateTime, 16, "StartTime");
                command.Parameters.Add("@SalesOrderId", SqlDbType.Int, 32, "SalesOrderId");
                command.Parameters.Add("@TotalSales", SqlDbType.Float, 32, "TotalSales");
            }
            catch (SqlException e)
            {
                Console.WriteLine(e.Message + e.StackTrace);
            }

            _adapterStopSignal = EventWaitHandle.OpenExisting(configinfo.AdapterStopSignal);
        }

        public override void Start()
        {
            ConsumeEvents();
        }

        public override void Resume()
        {
            ConsumeEvents();
        }

        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);
        }

        private void ConsumeEvents()
        {
            PointEvent<SalesByOrderId> currEvent;
            DequeueOperationResult result;

            try
            {
                while (true)
                {
                    if (AdapterState.Stopping == AdapterState)
                    {
                        // first, check if the adapter is being asked to stop. If yes, then
                        // make one attempt to dequeue the last event from the engine;
                        // do housekeeping before calling stopped;
                        // declare the adapter as stopped;
                        // signal to query that the adapter is stopped; and
                        // exit the worker thread
                        result = Dequeue(out currEvent);
                        PrepareToStop(currEvent, result);
                        Stopped();
                        _adapterStopSignal.Set();
                        return;
                    }

                    // NOTE: at any point during execution of the code below, if the Adapter state
                    // changes to Stopping, the engine will resume the adapter (i.e. call Resume())
                    // one more time, and the stopping condition will be trapped at the check above.

                    // Dequeue the event. If the dequeue fails, then the adapter state is suspended
                    // or stopping. Assume the former and do housekeeping, and call Ready() to indicate
                    // readiness to be resumed, and exit the worker thread.
                    result = Dequeue(out currEvent);
                    if (DequeueOperationResult.Empty == result)
                    {
                        PrepareToResume();
                        Ready();
                        return;
                    }
                    else
                    {
                        _eventsDequeued++;

                        command.Parameters["@EventKind"].Value = currEvent.EventKind.ToString();
                        command.Parameters["@StartTime"].Value = currEvent.StartTime;

                        if (EventKind.Insert == currEvent.EventKind)
                        {
                            command.Parameters["@SalesOrderId"].Value = currEvent.Payload.SalesOrderId;
                            command.Parameters["@TotalSales"].Value = currEvent.Payload.TotalSales;
                        }

                        command.ExecuteNonQuery();

                        PrintEvent(currEvent);
                        ReleaseEvent(ref currEvent);
                    }
                }
            }
            catch (AdapterException e1)
            {
                Console.WriteLine("ConsumeEvents - " + e1.Message + e1.StackTrace);
            }
            catch (SqlException e2)
            {
                Console.WriteLine("ConsumeEvents - " + e2.Message + e2.StackTrace);
            }
        }

        private void PrepareToStop(PointEvent<SalesByOrderId> currEvent, DequeueOperationResult result)
        {
            if (DequeueOperationResult.Success == result)
            {
                // Save/report info on the last dequeued event for later reconciliation with output stream
                // Console.WriteLine("Last dequeued event : " + _delimiter + CreateLineFromEvent(currEvent));

                // Release the event
                // NOTE: This is very important to avoid leaking memory allocated by the server for the event.
                ReleaseEvent(ref currEvent);
            }

            // release database resources
            connection.Close();
            connection.Dispose();
        }

        private void PrepareToResume()
        {
            // The server can asynchronously call Resume at any point in time while the execution control
            // is in this routine. This routine can be used for any housekeeping before being resumed.
        }

        private void PrintEvent(PointEvent<SalesByOrderId> evt)
        {
            if (EventKind.Insert == evt.EventKind)
            {
                Console.WriteLine("Output: " + "|INSERT|" + evt.StartTime + "|" +
                    evt.Payload.SalesOrderId + "|" + evt.Payload.TotalSales);
            }
            else
            {
                Console.WriteLine("Output: " + "|CTI|" + evt.StartTime + "|");
            }
        }
    }
}
