﻿namespace StreamInsight.Samples.OutputAdapters.TextFile
{
    using System;
    using System.Globalization;
    using System.IO;
    using System.Threading;
    using Microsoft.ComplexEventProcessing;
    using Microsoft.ComplexEventProcessing.Adapters;
    
    /// <summary>
    /// TextFileEdgeOutput adapter dequeues generic edge events into a CSV file with the format:
    ///   {EventKind}{delimiter}{EdgeType}{delimiter}{Starttime}{delimiter}{EndTime}{delimiter}{field2}{delimiter}...{fieldN}
    /// By generic events, we mean that, for a given output device (in this case a CSV file), you
    /// can specify an arbitrary payload class or struct at the time of query binding. This enables
    /// you to reuse this adapter implementation for multiple instantiations of the adapter - where
    /// each instantiation corresponds to a particular payload type, and is associated/bound 1:1
    /// with a query.
    /// <para/>
    /// Please study this code in conjunction with the adapter state transition diagram provided in
    /// the documentation.
    /// </summary>
    public class TextFileEdgeOutput : EdgeOutputAdapter
    {
        private StreamWriter _streamWriter;
        private string _line;
        private char _delimiter;

        private EventWaitHandle _adapterStopSignal;

        private CepEventType _bindtimeEventType;
        private int _eventsDequeued = 0;

        /// <summary>
        /// Initializes a new instance of the TextFileEdgeOutput class.
        /// <para/>
        /// See notes in TextFilePointOutput.cs
        /// </summary>
        /// <param name="configInfo"></param>
        /// <param name="cepEventType"></param>
        public TextFileEdgeOutput(TextFileOutputConfig configInfo, CepEventType cepEventType)
        {
            // If an empty string as filename was provided, dump output on the 
            // console
            _streamWriter = configInfo.OutputFileName.Length == 0 ?
                              new StreamWriter(Console.OpenStandardOutput()) :
                              new StreamWriter(configInfo.OutputFileName);
            
            _delimiter = configInfo.Delimiter;
            _bindtimeEventType = cepEventType;

            _adapterStopSignal = EventWaitHandle.OpenExisting(configInfo.AdapterStopSignal);
        }

        /// <summary>
        /// See notes in TextFilePointOutput.cs.
        /// </summary>
        public override void Start()
        {
            ConsumeEvents();
        }

        /// <summary>
        /// See notes in TextFilePointOutput.cs
        /// </summary>
        public override void Resume()
        {
            ConsumeEvents();
        }

        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);
        }

        /// <summary>
        /// Main driver to dequeue events and output them to the CSV file
        /// </summary>
        private void ConsumeEvents()
        {
            EdgeEvent lastDequeued = default(EdgeEvent);
            DequeueOperationResult result;

            try
            {
                while (true)
                {
                    if (AdapterState.Stopping == AdapterState)
                    {
                        result = Dequeue(out lastDequeued);
                        PrepareToStop(lastDequeued, result);
                        Stopped();

                        _adapterStopSignal.Set();

                        return;
                    }

                    result = Dequeue(out lastDequeued);

                    if (DequeueOperationResult.Empty == result)
                    {
                        PrepareToResume(lastDequeued);
                        Ready();
                        return;
                    }
                    else
                    {
                        _eventsDequeued++;
                        _line = CreateLineFromEvent(lastDequeued);
                        _streamWriter.WriteLine(_line);

                        // very important - release the event
                        ReleaseEvent(ref lastDequeued);
                    }
                }
            }
            catch (AdapterException e)
            {
                Console.WriteLine("ConsumeEvents - " + e.Message + e.StackTrace);
            }
        }

        private void PrepareToStop(EdgeEvent lastDequeued, DequeueOperationResult result)
        {
            if (DequeueOperationResult.Success == result)
            {
                // Save/report info on the last dequeued event for later 
                // reconciliation with output stream
                Console.WriteLine("Last dequeued event : " + _delimiter + CreateLineFromEvent(lastDequeued));

                // Release the event
                // NOTE: This is very important to avoid leaking memory 
                // allocated by the server for the event.
                ReleaseEvent(ref lastDequeued);
            }

            // release File resources
            _streamWriter.Flush();
            _streamWriter.Close();
            _streamWriter.Dispose();
        }

        private void PrepareToResume(EdgeEvent lastDequeued)
        {
            // The server can asynchronously call Resume at any point in time 
            // while the execution control is in this routine. This routine can
            // be used for any housekeeping before being resumed.
        }

        private string CreateLineFromEvent(EdgeEvent lastDequeued)
        {
            string line = default(string);

            if (EventKind.Cti == lastDequeued.EventKind)
            {
                line += "CTI" + _delimiter + lastDequeued.StartTime.ToString();
            }
            else
            {
                line += "INSERT" + _delimiter + lastDequeued.EdgeType.ToString() + _delimiter;
                line += lastDequeued.StartTime.ToString() + _delimiter +
                    ((EdgeType.End == lastDequeued.EdgeType) ? lastDequeued.EndTime.ToString() : string.Empty) + _delimiter;

                for (int ordinal = 0; ordinal < _bindtimeEventType.FieldsByOrdinal.Count; ordinal++)
                {
                    CepEventTypeField evtField = _bindtimeEventType.FieldsByOrdinal[ordinal];
                    object value = Convert.ChangeType(lastDequeued.GetField(ordinal), evtField.Type.ClrType, CultureInfo.CurrentCulture);
                    line += value != null ? value.ToString() : "NULL";
                    line += _delimiter;
                }
            }

            return line;
        }
    }
}