﻿namespace StreamInsight.Samples.OutputAdapters.TextFile
{
    using System;
    using System.Globalization;
    using System.IO;
    using System.Threading;
    using Microsoft.ComplexEventProcessing;
    using Microsoft.ComplexEventProcessing.Adapters;

    /// <summary>
    /// Initializes a new instance of the class TextFileIntervalOutput.
    /// <para/>
    /// TextFileIntervalOutput adapter dequeues generic interval events into a CSV file with the format:
    ///   {EventKind}{delimiter}{Starttime}{delimiter}{EndTime}{delimiter}{field2}{delimiter}...{fieldN}
    /// By generic events, we mean that, for a given output device (in this case a CSV file), you
    /// can specify an arbitrary payload class or struct at the time of query binding. This enables
    /// you to reuse this adapter implementation for multiple instantiations of the adapter - where
    /// each instantiation corresponds to a particular payload type, and is associated/bound 1:1
    /// with a query.
    /// <para/>
    /// Please study this code in conjunction with the adapter state transition diagram provided in
    /// the documentation.
    /// </summary>
    public class TextFileIntervalOutput : IntervalOutputAdapter
    {
        private StreamWriter _streamWriter;
        private string _line;
        private char _delimiter;

        private EventWaitHandle _adapterStopSignal;

        private CepEventType _bindtimeEventType;
        private int _eventsDequeued = 0;

        /// <summary>
        /// Initializes a new instance of the TextFileIntervalOutput class.
        /// Notes on the Constructor Method:
        /// See TextFilePointOutput.cs
        /// </summary>
        /// <param name="configInfo"></param>
        /// <param name="cepEventType"></param>
        public TextFileIntervalOutput(TextFileOutputConfig configInfo, CepEventType cepEventType)
        {
            // If an empty string as filename was provided, dump output on the 
            // console
            _streamWriter = configInfo.OutputFileName.Length == 0 ?
                              new StreamWriter(Console.OpenStandardOutput()) :
                              new StreamWriter(configInfo.OutputFileName);

            _delimiter = configInfo.Delimiter;
            _bindtimeEventType = cepEventType;

            _adapterStopSignal = EventWaitHandle.OpenExisting(configInfo.AdapterStopSignal);
        }

        /// <summary>
        /// See notes in TextFilePointOutput.cs
        /// </summary>
        public override void Start()
        {
            ConsumeEvents();
        }

        /// <summary>
        /// See notes in TextFilePointOutput.cs
        /// </summary>
        public override void Resume()
        {
            ConsumeEvents();
        }

        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);
        }

        /// <summary>
        /// Main driver to dequeue events and output them to the CSV file
        /// See notes and comments in TextFilePointOutput.cs
        /// </summary>
        private void ConsumeEvents()
        {
            IntervalEvent currEvent = default(IntervalEvent);
            DequeueOperationResult result;

            try
            {
                while (true)
                {
                    if (AdapterState.Stopping == AdapterState)
                    {
                        result = Dequeue(out currEvent);
                        PrepareToStop(currEvent, result);
                        Stopped();
                        _adapterStopSignal.Set();
                        return;
                    }

                    result = Dequeue(out currEvent);
                    if (DequeueOperationResult.Empty == result)
                    {
                        PrepareToResume(currEvent);
                        Ready();
                        return;
                    }
                    else
                    {
                        _eventsDequeued++;
                        _line = CreateLineFromEvent(currEvent);
                        _streamWriter.WriteLine(_line);

                        // very important - release the event
                        ReleaseEvent(ref currEvent);
                    }
                }
            }
            catch (AdapterException e)
            {
                Console.WriteLine("ConsumeEvents - " + e.Message + e.StackTrace);
            }
        }

        private void PrepareToStop(IntervalEvent currEvent, DequeueOperationResult result)
        {
            if (DequeueOperationResult.Success == result)
            {
                // Save/report info on the last dequeued event for later reconciliation with output stream
                // Console.WriteLine("Last dequeued event : " + _delimiter + CreateLineFromEvent(currEvent));

                // Release the event
                // NOTE: This is very important to avoid leaking memory allocated by the server for the event.
                ReleaseEvent(ref currEvent);
            }

            // release File resources
            _streamWriter.Flush();
            _streamWriter.Close();
            _streamWriter.Dispose();
        }

        private void PrepareToResume(IntervalEvent currEvent)
        {
            // The server can asynchronously call Resume at any point in time while the execution control
            // is in this routine. This routine can be used for any housekeeping before being resumed.
        }

        private string CreateLineFromEvent(IntervalEvent currEvent)
        {
            string line = default(string);

            if (EventKind.Cti == currEvent.EventKind)
            {
                line += " CTI" + _delimiter + currEvent.StartTime;
            }
            else
            {
                line += "INSERT" + _delimiter + currEvent.StartTime + _delimiter + currEvent.EndTime + _delimiter;
                for (int ordinal = 0; ordinal < _bindtimeEventType.FieldsByOrdinal.Count; ordinal++)
                {
                    CepEventTypeField evtField = _bindtimeEventType.FieldsByOrdinal[ordinal];
                    object value = Convert.ChangeType(currEvent.GetField(ordinal), evtField.Type.ClrType, CultureInfo.CurrentCulture);
                    line += value != null ? value.ToString() : "NULL";
                    line += _delimiter;
                }
            }

            return line;
        }
    }
}
