﻿namespace StreamInsight.Samples.OutputAdapters.TextFile
{
    using System;
    using System.Globalization;
    using System.IO;
    using System.Threading;
    using Microsoft.ComplexEventProcessing;
    using Microsoft.ComplexEventProcessing.Adapters;

    /// <summary>
    /// TextFilePointOutput adapter dequeues generic point events into a CSV file with the format:
    ///   {EventKind}{delimiter}{StartTime}{delimiter}{field1}{delimiter}...{fieldN}
    /// By generic events, we mean that, for a given output device (in this case a CSV file), you
    /// can specify an arbitrary payload class or struct at the time of query binding. This enables
    /// you to reuse this adapter implementation for multiple instantiations of the adapter - where
    /// each instantiation corresponds to a particular payload type, and is associated/bound 1:1
    /// with a query.
    /// <para/>
    /// Please study this code in conjunction with the adapter state transition diagram provided in
    /// the documentation.
    /// </summary>
    public class TextFilePointOutput : PointOutputAdapter
    {
        private StreamWriter _streamWriter;
        private string _line;
        private char _delimiter;
        private EventWaitHandle _adapterStopSignal;
        private CepEventType _bindtimeEventType;
        private int _eventsDequeued = 0;

        /// <summary>
        /// Initializes a new instance of the TextFilePointOutput class.
        /// <para/>
        /// Notes on the Constructor Method:
        /// TextFileOutputConfig is provided via the TextFileOutputFactory class, and contains the following
        /// configuration information:
        ///            Filename, Delimiter
        /// The filename is a parameter to the StreamWriter method - which will overwrite an existing file,
        /// or create if non-existent. If given an emtpy string, the adapter will just dump to the console.
        /// The delimiter is a single character.
        /// <para/>
        /// CepEventType represents the generic point event type that will be passed into this constructor
        /// from the query binding via the adapter factory class. The payload class or struct supports several
        /// CLR types. This example can handle most scalar fields.The individual fields are combined into a
        /// string and written as a line using synchronous writes.
        /// <para/>
        /// The EventWaitHandle is a synchronization between the output adapter instance and the query
        /// (See the query snippet and MSDN documentation for usage of this class).
        /// </summary>
        /// <param name="configInfo"></param>
        /// <param name="cepEventType"></param>
        public TextFilePointOutput(TextFileOutputConfig configInfo, CepEventType cepEventType)
        {
            // If an empty string as filename was provided, dump output on the 
            // console
            _streamWriter = configInfo.OutputFileName.Length == 0 ?
                              new StreamWriter(Console.OpenStandardOutput()) :
                              new StreamWriter(configInfo.OutputFileName);

            _delimiter = configInfo.Delimiter;
            _bindtimeEventType = cepEventType;
            
            _adapterStopSignal = EventWaitHandle.OpenExisting(configInfo.AdapterStopSignal);
        }

        /// <summary>
        /// Start is the first method to be invoked by the CEP engine once the output adapter has been instantiated.
        /// So any initializations that cannot be covered in the constructor can be placed here. Start() is called
        /// in a separate worker thread initiated by the engine. In this example (and in most scenarios), Start
        /// begins consuming events once it is invoked.
        /// </summary>
        public override void Start()
        {
            ConsumeEvents();
        }

        /// <summary>
        /// Resume is invoked by the engine once it returns from being scheduled away from Start, and only after
        /// the adapter has called Ready(). Resume continues to consume events.
        /// </summary>
        public override void Resume()
        {
            ConsumeEvents();
        }

        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);
        }

        /// <summary>
        /// Main driver to dequeue events and output them to the CSV file
        /// </summary>
        private void ConsumeEvents()
        {
            PointEvent currEvent = default(PointEvent);
            DequeueOperationResult result;

            try
            {
                while (true)
                {
                    if (AdapterState.Stopping == AdapterState)
                    {
                        // first, check if the adapter is being asked to stop. If yes, then
                        // make one attempt to dequeue the last event from the engin;
                        // do housekeeping before calling stopped;
                        // declare the adapter as stopped;
                        // signal to query that the adapter is stopped; and
                        // exit the worker thread
                        result = Dequeue(out currEvent);
                        PrepareToStop(currEvent, result);
                        Stopped();
                        _adapterStopSignal.Set();
                        return;
                    }

                    // NOTE: at any point during execution of the code below, if the Adapter state
                    // changes to Stopping, the engine will resume the adapter (i.e. call Resume())
                    // one more time, and the stopping condition will be trapped at the check above.

                    // Dequeue the event. If the dequeue fails, then the adapter state is suspended
                    // or stopping. Assume the former and do housekeeping, and call Ready() to indicate
                    // readiness to be resumed, and exit the worker thread.
                    result = Dequeue(out currEvent);
                    if (DequeueOperationResult.Empty == result)
                    {
                        PrepareToResume();
                        Ready();
                        return;
                    }
                    else
                    {
                        _eventsDequeued++;
                        _line = CreateLineFromEvent(currEvent);
                        _streamWriter.WriteLine(_line);

                        // very important - release the event
                        ReleaseEvent(ref currEvent);
                    }
                }
            }
            catch (AdapterException e)
            {
                Console.WriteLine("ConsumeEvents - " + e.Message + e.StackTrace);
            }
        }

        private void PrepareToStop(PointEvent currEvent, DequeueOperationResult result)
        {
            if (DequeueOperationResult.Success == result)
            {
                // Save/report info on the last dequeued event for later reconciliation with output stream
                // Console.WriteLine("Last dequeued event : " + _delimiter + CreateLineFromEvent(currEvent));

                // Release the event
                // NOTE: This is very important to avoid leaking memory allocated by the server for the event.
                ReleaseEvent(ref currEvent);
            }

            // release File resources
            _streamWriter.Flush();
            _streamWriter.Close();
            _streamWriter.Dispose();
        }

        private void PrepareToResume()
        {
            // The server can asynchronously call Resume at any point in time while the execution control
            // is in this routine. This routine can be used for any housekeeping before being resumed.
        }

        // Create line from event of format:
        // Line format: <EdgeType>,<StartTime>,<EndTime>,<comma separated Payload Fields>
        private string CreateLineFromEvent(PointEvent currEvent)
        {
            string line = default(string);
            
            if (EventKind.Cti == currEvent.EventKind)
            {
                line += " CTI" + _delimiter + currEvent.StartTime;
            }
            else
            {
                line += " INSERT" + _delimiter + currEvent.StartTime + _delimiter;
                for (int ordinal = 0; ordinal < _bindtimeEventType.FieldsByOrdinal.Count; ordinal++)
                {
                    CepEventTypeField evtField = _bindtimeEventType.FieldsByOrdinal[ordinal];
                    object value = Convert.ChangeType(currEvent.GetField(ordinal), evtField.Type.ClrType, CultureInfo.CurrentCulture);
                    line += (value != null) ? value.ToString() : "NULL";
                    line += _delimiter;
                }
            }

            return line;
        }
    }
}