﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Microsoft.ComplexEventProcessing;
using Microsoft.ComplexEventProcessing.Adapters;
using System.Globalization;
using StreamInsight.Samples.OutputAdapters.Trace;

namespace StreamInsight.Samples.OutputAdapters.Trace
{
    /// <summary>
    /// 
    /// </summary>
    public class TracePointOutput : PointOutputAdapter
    {     
        private int _eventsDequeued = 0;

        /// <summary>
        /// The CEP event type - we find out what our incoming event type is at 
        /// "bind time".
        /// </summary>
        private CepEventType _bindtimeEventType;

        /// <summary>
        /// The run-time configuration object.
        /// </summary>
        private TraceConfig _config;

        /// <summary>
        /// Create an instance of the specified trace output adapater, based on:
        /// - Payload type (CepEventType)
        /// - Configuration information (output stream type)
        /// </summary>
        public TracePointOutput(TraceConfig config, CepEventType type)
        {
            _bindtimeEventType = type;
            _config = config;
        }

        public override void Resume()
        {
            ConsumeEvents();
        }

        public override void Start()
        {
            ConsumeEvents();
        }

        /// <summary>
        /// Main worker thread function responsible for dequeueing events and 
        /// posting them to the output stream.
        /// </summary>
        private void ConsumeEvents()
        {
            PointEvent currEvent = default(PointEvent);
            DequeueOperationResult result;

            // Get a reference to the output stream handler, based on the current 
            // trace type.
            PrintHandler outHandler = TraceRouter.GetHandler(_config.TraceType,
                _config.TraceName);

            try
            {
                while (true)
                {
                    if (AdapterState.Stopping == AdapterState)
                    {
                        // first, check if the adapter is being asked to stop. If yes, then
                        // make one attempt to dequeue the last event from the engin;
                        // do housekeeping before calling stopped;
                        // declare the adapter as stopped;
                        // signal to query that the adapter is stopped; and
                        // exit the worker thread
                        result = Dequeue(out currEvent);
                        PrepareToStop(currEvent, result);
                        Stopped();
                        return;
                    }

                    // NOTE: at any point during execution of the code below, if the Adapter state
                    // changes to Stopping, the engine will resume the adapter (i.e. call Resume())
                    // one more time, and the stopping condition will be trapped at the check above.

                    // Dequeue the event. If the dequeue fails, then the adapter state is suspended
                    // or stopping. Assume the former and do housekeeping, and call Ready() to indicate
                    // readiness to be resumed, and exit the worker thread.
                    result = Dequeue(out currEvent);
                    if (DequeueOperationResult.Empty == result)
                    {
                        PrepareToResume();
                        Ready();
                        return;
                    }
                    else
                    {
                        _eventsDequeued++;
                   
                        if (currEvent.EventKind == EventKind.Insert ||
                            (currEvent.EventKind == EventKind.Cti && 
                             _config.DisplayCtiEvents))
                        {
                            outHandler(FormatEventForDisplay(currEvent));
                        }

                        // very important - release the event
                        ReleaseEvent(ref currEvent);
                    }
                }
            }
            catch (Exception ex0)
            {
                Debug.WriteLine(ex0.ToString());
            }
        }

        private void PrepareToStop(PointEvent currEvent, DequeueOperationResult result)
        {
            if (DequeueOperationResult.Success == result)
            {
                // Release the event - NOTE: This is very important to avoid leaking 
                // memory allocated by the server for the event.                
                ReleaseEvent(ref currEvent);
            }
        }

        /// <summary>
        /// The server can asynchronously call Resume at any point in time while the execution control
        /// is in this routine. This routine can be used for any housekeeping before being resumed.
        /// </summary>
        private void PrepareToResume()
        {
            
        }

        /// <summary>
        /// Given an untyped PointEvent generate a textual representation suitable for console,
        /// trace, etc, output.
        /// </summary>
        /// <param name="currEvent">An untyped PointEvent</param>
        /// <returns>A formatted string of the contents of the PointEvent</returns>
        private string FormatEventForDisplay(PointEvent currEvent)
        {
            StringBuilder sb = new StringBuilder();

            if (EventKind.Cti == currEvent.EventKind)
            {
                sb.AppendFormat(" CTI at {0}", currEvent.StartTime);
            }
            else
            {
                sb.AppendFormat("Point Event at {0}:\r\n", currEvent.StartTime);
                for (int ordinal = 0; ordinal < _bindtimeEventType.FieldsByOrdinal.Count; ordinal++)
                {
                    CepEventTypeField evtField = _bindtimeEventType.FieldsByOrdinal[ordinal];
                    object value = Convert.ChangeType(currEvent.GetField(ordinal), evtField.Type.ClrType, CultureInfo.CurrentCulture);
                    sb.AppendFormat("\t{0} = {1}\r\n", evtField.Name, ((value != null) ? value.ToString() : "NULL"));
                }
            }
            return sb.ToString();
        }

    }
}
