﻿using System;
using System.Collections.Generic;
using System.IO;

namespace StreamInsight.Samples.OutputAdapters.Trace
{
    /// <summary>
    /// Delegate type for output streams (used to abstract the details of 
    /// pushing data to the output stream)
    /// </summary>
    public delegate void PrintHandler(string msg);

    internal abstract class TraceRouter
    {
        public static PrintHandler GetHandler(string type, string traceName)
        {
            if (type == TraceTypeValue.File.ToString())
            {
                FileTracer ft = new FileTracer(traceName);
                return ft.WriteLine;
            }
            return streamMap[type];
        }

        public static PrintHandler GetHandler(TraceTypeValue type, string traceName)
        {
            if (type == TraceTypeValue.File)
            {
                FileTracer ft = new FileTracer(traceName);
                return ft.WriteLine;
            }
            return streamMap[type.ToString()];
        }

        #region "Output Stream Handlers"

        
        /// <summary>
        /// Map from the trace type name to the output stream method that can accept
        /// output messages.
        /// </summary>
        private static Dictionary<string, PrintHandler> streamMap =
            new Dictionary<string, PrintHandler>
            {
                { TraceTypeValue.Console.ToString(),   Console.WriteLine   },
                { TraceTypeValue.Trace.ToString(),     WriteMessageToTrace },
                { TraceTypeValue.Debug.ToString(),     WriteMessageToDebug },                                
            };
        
        /// <summary>
        /// Wrap Trace.WriteLine in a delegate due to the [Conditional] attribute
        /// (cannot define a static delegate on a conditional method)
        /// </summary>
        private static void WriteMessageToTrace(string msg)
        {
            System.Diagnostics.Trace.WriteLine(msg);
        }

        /// <summary>
        /// Wrap Trace.WriteLine in a delegate due to the [Conditional] attribute
        /// (cannot define a static delegate on a conditional method)
        /// </summary>
        private static void WriteMessageToDebug(string msg)
        {
            System.Diagnostics.Debug.WriteLine(msg);
        }
        #endregion            
    }

    internal class FileTracer
    {
        private string _fileName;

        public FileTracer(string fileName)
        {
            _fileName = fileName;
        }

        public void WriteLine(string msg)
        {
            using (TextWriter tw = new StreamWriter(_fileName, true))
                tw.WriteLine(msg);               
        }
    }
}
