﻿namespace StreamInsight.Samples.PatternDetector
{
    using System.Collections.Generic;
    using StreamInsight.Samples.Extensibility.PatternUDO;

    // The input is a sequence of stock price ticks.
    // We define an ATN that looks for a sequence of downticks, immediately 
    // followed by an equal number of upticks.
    // In this case, the register (counter) keeps track of '#downticks - #upticks', and when 
    // this reaches 0, we have found the pattern.

    class ATN_EqualDownticksUpticks : IATNDescriptorGenerator<StockTick, ATNRegisterType>
    {
        public ATNDescriptor<StockTick, ATNRegisterType> GetATNDescriptor()
        {
            ATNDescriptor<StockTick, ATNRegisterType> atnDescriptor =
                new ATNDescriptor<StockTick, ATNRegisterType>();

            // Specify the start and final states
            atnDescriptor.StartState = 0;
            atnDescriptor.AddFinalState(2);

            // Specify the arcs in the automaton

            // If we see a downtick in the start state 0, we make 
            // two transitions: one stays in state 0 and the other
            // progresses the automaton to state 1. In both cases,
            // the counter is incremented by 1.
            atnDescriptor.AddArc(0, 0, Transition0);
            atnDescriptor.AddArc(0, 1, Transition0);

            // If we see an uptick, and the counter is greater than
            // 1, we decrement the counter and stay in the same state
            // (state 1) because the pattern has not yet been found
            atnDescriptor.AddArc(1, 1, Transition1);

            // If we see an uptick, and the counter is equal to 1, we 
            // decrement the counter (so it becomes 0) and move to state 2,
            // which is a final state and produces a pattern match output
            atnDescriptor.AddArc(1, 2, Transition2);

            // Specify the default (initial) register content
            atnDescriptor.DefaultRegister = new ATNRegisterType { counter = 0 };

            return atnDescriptor;
        }

        // Define the transition functions used in the ATN (see explanation in 
        // earlier comments)
        public static bool Transition0(List<StockTick> inputEvents, ATNRegisterType oldRegister, out ATNRegisterType newRegister)
        {
            // If there are many events with same timestamp, we simply look at the first event
            // and ignore the rest (this will not happen if each stock tick for the same stock
            // has a unique timestamp)
            if (inputEvents[0].IsDowntick())
            {
                newRegister = new ATNRegisterType { counter = oldRegister.counter + 1 };
                return true;
            }
            newRegister = oldRegister;
            return false;
        }

        public static bool Transition1(List<StockTick> inputEvents, ATNRegisterType oldRegister, out ATNRegisterType newRegister)
        {
            if ((inputEvents[0].IsUptick()) && (oldRegister.counter > 1))
            {
                newRegister = new ATNRegisterType { counter = oldRegister.counter - 1 };
                return true;
            }
            newRegister = oldRegister;
            return false;
        }

        public static bool Transition2(List<StockTick> inputEvents, ATNRegisterType oldRegister, out ATNRegisterType newRegister)
        {
            if ((inputEvents[0].IsUptick()) && (oldRegister.counter == 1))
            {
                newRegister = new ATNRegisterType { counter = oldRegister.counter - 1 };
                return true;
            }
            newRegister = oldRegister;
            return false;
        }
    }
}
