﻿namespace StreamInsight.Samples.SqlApp
{
    using System;
    using System.Threading;
    using Microsoft.ComplexEventProcessing;
    using Microsoft.ComplexEventProcessing.Linq;
    using StreamInsight.Samples.EventTypes;
    using StreamInsight.Samples.InputAdapters.Sql;
    using StreamInsight.Samples.OutputAdapters.Sql;

    public class SqlApp
    {
        private const int OnceEveryFiveEvents = 5;

        public static void Main()
        {
            //BB: Added these lines
            Server server = Server.Create("default");
            Application application = server.CreateApplication("ImplicitServer");

            // Set up device configuration info for database input and output tables
            // Please read comments in SqlTypeInput.cs and SqlTypedOutput.cs
            SqlInputConfig inputConfig = new SqlInputConfig();
            inputConfig.CtiFrequency = OnceEveryFiveEvents;
            inputConfig.ConnectionString = @"integrated security = true; database = AdventureWorks2008";

            SqlOutputConfig outputConfig = new SqlOutputConfig();
            outputConfig.AdapterStopSignal = "StopAdapter";
            outputConfig.ConnectionString = @"integrated security = true; database = AdventureWorks2008";

            // Define the input stream from the adapter. The factory creates the point adapter instance
            // based on the event shape specification (see SqlInputFactory.cs)
            CepStream<SalesOrderDetail> producer = CepStream<SalesOrderDetail>.Create(
                "TypedSqlInPointOutPoint", 
                typeof(SqlInputFactory),                
                inputConfig,
                EventShape.Point);

            // Specify the query template - the query groups each sales order by id and
            // provides the total sales for each sales order. 
            // BB: Same SnapshotWindow
            CepStream<SalesByOrderId> sampleProducer =
                from e in producer.AlterEventDuration(e => TimeSpan.FromMinutes(1))
                group e by e.SalesOrderId into eachSalesGroup
                from win in eachSalesGroup.SnapshotWindow(SnapshotWindowOutputPolicy.Clip)
               select new SalesByOrderId { SalesOrderId = eachSalesGroup.Key, TotalSales = win.Sum(e => e.UnitPrice) };

            // Define the consumer of events
            // BB: Needs QueryName and description
            Query query = sampleProducer.ToQuery(application,
                "SampleQuery",
                "Description of the Sample Query",
                typeof(SqlOutputFactory),
                outputConfig,
                EventShape.Point,
                StreamEventOrder.ChainOrdered);

            // The adapters work asynchronously with the main thread. If the source has a known
            // end to the event stream (end of table, for example), you may want to stop the query
            // only after you have known that all the events have been written into the sink.
            // Otherwise, this main thread will  stop the worker thread while ProduceEvents
            // is going on.
            EventWaitHandle adapterStopSignal =
                new EventWaitHandle(false, EventResetMode.ManualReset, outputConfig.AdapterStopSignal);

            // Start the query
            query.Start();

            if (!adapterStopSignal.WaitOne())
            {
                Console.WriteLine("Waiting for signal from output adapter");
            }

            query.Stop();
            Console.WriteLine("Query and Program Stopped!");
        }
    }
}
