﻿namespace StreamInsight.Samples.UserAggregateQuery
{
    using System;
    using System.Diagnostics;
    using Microsoft.ComplexEventProcessing;
    using Microsoft.ComplexEventProcessing.Linq;
    using StreamInsight.Samples.Extensibility.TimeWeightedAverage;
    using StreamInsight.Samples.InputAdapters.DataSimulator;
    using StreamInsight.Samples.OutputAdapters.Trace;

    public class SampleReading : IRandomInit
    {
        public string SourceId { get; set; }
        public double Value { get; set; }

        /// <summary>
        /// When invoked, assign a random temperature value between 0 and 100
        /// to the Value property.
        /// </summary>
        public void Init()
        {
            SourceId = "TempSensor1";
            Value = new Random().Next(100);
        }
    }

    class UserAggregateQuery
    {
        static void Main(string[] args)
        {
            Trace.Listeners.Add(new TextWriterTraceListener(Console.Out));

            var inputConfData = new SimulatedDataInputConfig
            {
                CtiFrequency = 1,
                EventInterval = 1000
            };

            var outputConf = new TraceConfig
            {
                DisplayCtiEvents = true,
                TraceName = "output",
                TraceType = TraceTypeValue.Console
            };

            try
            {
                Trace.WriteLine("Create event producer");
                var inputDataStream = CepStream<SampleReading>.Create("DataInput", typeof(SimulatedDataInputFactory), inputConfData, EventShape.Interval);

                // Run the user-defined aggregate on the Value field every three seconds.
                // A configuration is used to demonstrate the ability to
                // pass parameters to the UDA.
                // BB: New Parameter on Constructor
                var twaStream = from win in inputDataStream.TumblingWindow(TimeSpan.FromSeconds(3), HoppingWindowOutputPolicy.ClipToWindowEnd)
                                select new { avg = win.twa(e => e.Value, new TwaConfig() { weight = 1 }) };

                Trace.WriteLine("Create event consumer");

                // BB: Added name and description
                Query query = twaStream.ToQuery("QueryName", "QueryDescription", typeof(TraceOutputFactory), outputConf, EventShape.Interval, StreamEventOrder.FullyOrdered);

                Trace.WriteLine("Start query (stop with enter)");

                query.Start();
                Console.ReadLine();
                query.Stop();
            }
            catch (Exception e)
            {
                Trace.WriteLine(e.ToString());
            }

            Console.WriteLine("\npress enter to exit application");
            Console.ReadLine();
        }
    }
}
