namespace StreamInsight.Samples.Extensibility.PatternUDO
{
    using System;
    using System.Collections.Generic;
    using Microsoft.ComplexEventProcessing;
    using Microsoft.ComplexEventProcessing.Extensibility;

    internal class EventSet<TInput>
    {
        private DateTimeOffset _startTime = DateTimeOffset.MaxValue;
        public DateTimeOffset StartTime
        {
            get { return _startTime; }
        }

        private DateTimeOffset _endTime = DateTimeOffset.MaxValue;
        public DateTimeOffset EndTime
        {
            get { return _endTime; }
        }

        private List<IntervalEvent<TInput>> _eventList = new List<IntervalEvent<TInput>>();
        public List<IntervalEvent<TInput>> EventList
        {
            get { return _eventList; }
        }


        public void Add(IntervalEvent<TInput> evt)
        {
            _eventList.Add(evt);
            if (evt.StartTime < _startTime) _startTime = evt.StartTime;
            if (evt.EndTime < _endTime) _endTime = evt.EndTime;
        }
    };

    public class ATNCore<TInput, TRegister> : CepTimeSensitiveOperator<TInput, TRegister>
    {
        ATNDescriptor<TInput, TRegister> _atnDescriptor = null;
        WindowDescriptor _windowDescriptor;
        LinkedList<Run> _maxRuns = null;
        List<IntervalEvent<TRegister>> outputList = new List<IntervalEvent<TRegister>>();

        public ATNCore(ATNDescriptor<TInput, TRegister> atnDescriptor)
        {
            _atnDescriptor = atnDescriptor;
        }

        public ATNCore(string typeName)
        {
            // Load the ATN descriptor using the fully qualified type name
            Type _inputType = Type.GetType(typeName, true);

            // try to get the type of the factory interface
            Type factoryInterface = null;
            foreach (Type implementedInterface in _inputType.GetInterfaces())
            {
                    if (implementedInterface.IsGenericType &&
                        (implementedInterface.GetGenericTypeDefinition() == typeof(IATNDescriptorGenerator<,>)))
                    {
                        // check to see that we have only one factory interface 
                        if (factoryInterface != null)
                        {
                            throw new InvalidOperationException("Supplied assembly implements more than one IATNDescriptorGenerator interface");
                        }
                        factoryInterface = implementedInterface;
                     }
            }

            // The object must implement at least one factory interface
            if (null == factoryInterface)
            {
                throw new Exception("Unable to find IATNDescriptorGenerator interface");
            }

            if ( (factoryInterface.GetGenericArguments().Length==2) &&
                 (factoryInterface.GetGenericArguments()[0] == typeof(TInput)) &&
                 (factoryInterface.GetGenericArguments()[1] == typeof(TRegister)))
            {

                IATNDescriptorGenerator<TInput, TRegister> atnDescriptorGenerator = (IATNDescriptorGenerator<TInput, TRegister>)Activator.CreateInstance(_inputType);
                _atnDescriptor = atnDescriptorGenerator.GetATNDescriptor();
            }
            else
            {
                throw new InvalidOperationException("Supplied IATNDescriptorGenerator has invalid input generic types");
            }

            if (_atnDescriptor == null)
            {
                throw new Exception("Specfied ATN descriptor could not be loaded");
            }
        }


        public override IEnumerable<IntervalEvent<TRegister>> GenerateOutput(IEnumerable<IntervalEvent<TInput>> intervalEvents, WindowDescriptor windowDescriptor)
        {
            _windowDescriptor = windowDescriptor;
            _maxRuns = null;
            outputList.Clear();
            
            SortedDictionary<DateTimeOffset, EventSet<TInput>> allEvents = new SortedDictionary<DateTimeOffset,EventSet<TInput>>();

            foreach (IntervalEvent<TInput> evt in intervalEvents)
            {
                if (evt.StartTime == _windowDescriptor.StartTime)
                {
                    // Assume for now that this condition implies a clipped event; hence, we discard the event
                    // TODO: In future, when we receive unclipped events, we need to remove this check
                    continue;
                }

                EventSet<TInput> e;
                if (! allEvents.TryGetValue(evt.StartTime, out e)) {
                    e = new EventSet<TInput>();
                    allEvents.Add(evt.StartTime, e);
                }
                e.Add(evt);
            }

            foreach (var item in allEvents)
            {
                Insert(item.Value);
            }

            List<IntervalEvent<TRegister>> outputCollection = new List<IntervalEvent<TRegister>>();
            foreach (IntervalEvent<TRegister> treg in outputList)
            {
                outputCollection.Add(treg);
            }
            return outputCollection;
        }

        private struct Run
        {
            private DateTimeOffset _startEventVs;
            private DateTimeOffset _startEventVe;
            private EventSet<TInput> _endEvent;
            
            internal TRegister _register;
            internal int _currentState;

            internal Run(EventSet<TInput> startEvent, EventSet<TInput> endEvent, int currentState, TRegister register)
            {
                _startEventVs = startEvent.StartTime;
                _startEventVe = startEvent.EndTime;
                _endEvent = endEvent;
                _register = register;
                _currentState = currentState;
            }

            internal Run(Run oldRun, EventSet<TInput> endEvent, int currentState, TRegister register)
            {
                _startEventVs = oldRun._startEventVs;
                _startEventVe = oldRun._startEventVe;
                _endEvent = endEvent;
                _register = register;
                _currentState = currentState;
            }
        }

        private void ApplyTransition(EventSet<TInput> inputEvents, ref Run r, ref LinkedList<Run> newRuns)
        {
            DateTimeOffset Vs = inputEvents.StartTime;

            int fromState = r._currentState;

            List<int> toStates = _atnDescriptor.GetToStates(fromState);

            foreach (int toState in toStates)
            {
                TRegister targetRegister;
                if (_atnDescriptor.ApplyTransition(inputEvents, fromState, toState, r._register, out targetRegister))
                {
                        Run newRun = new Run(r, inputEvents, toState, targetRegister);

                        if (_atnDescriptor.IsFinalState(toState))
                        {
                            IntervalEvent<TRegister> outputEvent = CreateIntervalEvent();
                            outputEvent.Payload = newRun._register;
                            outputEvent.StartTime = inputEvents.StartTime;
                            outputEvent.EndTime = _windowDescriptor.EndTime;
                            outputList.Add(outputEvent);
                        }
                        newRuns.AddLast(newRun);
                    }
            }
        }

        private void ApplyTransitions(EventSet<TInput> e, LinkedList<Run> oldRuns, ref LinkedList<Run> newRuns)
        {
            foreach (Run r in oldRuns)
            {
                Run rr = r;
                ApplyTransition(e, ref rr, ref newRuns);
            }
        }

        private void Insert(EventSet<TInput> eventSet)
        {
            // Insert assumes that EventSets are fed in increasing order of timestamps

            LinkedList<Run> newRuns = new LinkedList<Run>();

            // Step 1: Create new runs from existing runs that end at immediately preceeding sequence number
            LinkedList<Run> currRuns = _maxRuns;
            if (currRuns != null)
            {
                ApplyTransitions(eventSet, currRuns, ref newRuns);
            }

            // Step 2: Create new runs starting from the new sequence number
            Run r = new Run(eventSet, eventSet, _atnDescriptor.StartState, _atnDescriptor.DefaultRegister);
            ApplyTransition(eventSet, ref r, ref newRuns);

            if (newRuns.Count == 0) newRuns = null;
           _maxRuns = newRuns;
        }
    };
    
}
